@@ -15,6 +15,8 @@ from ._exceptions import (
     InvalidInputError,
     InternalError,
     DialogNotActiveError,
+    QTrioWarning,
+    ApplicationQuitWarning,
 )
 
 from ._core import (
@@ -29,7 +31,6 @@ from ._core import (
     registered_event_type,
     register_event_type,
     register_requested_event_type,
-    Reenter,
 )
 
 from ._qt import Signal
