@@ -4,29 +4,33 @@ Attributes:
     _reenter_event_type: The event type enumerator for our reenter events.
 """
 import contextlib
-import functools
 import math
 import sys
-import traceback
 import typing
+import typing_extensions
+import warnings
 
 import async_generator
 import attr
 import outcome
-from qtpy import QtCore
-from qtpy import QtGui
-from qtpy import QtWidgets
+import qts
 import trio
 import trio.abc
 
 import qtrio
 import qtrio._qt
-import qtrio._util
-
-_reenter_event_type: typing.Optional[QtCore.QEvent.Type] = None
 
 
-def registered_event_type() -> typing.Optional[QtCore.QEvent.Type]:
+if typing.TYPE_CHECKING or "sphinx_autodoc_typehints" in sys.modules:
+    from qts import QtCore
+    from qts import QtGui
+    from qts import QtWidgets
+
+
+_reenter_event_type: typing.Optional["QtCore.QEvent.Type"] = None
+
+
+def registered_event_type() -> typing.Optional["QtCore.QEvent.Type"]:
     """Get the registered event type.
 
     Returns:
@@ -45,6 +49,8 @@ def register_event_type() -> None:
         qtrio.EventTypeRegistrationFailedError:  if a type was not able to be
             registered.
     """
+    from qts import QtCore
+
     global _reenter_event_type
 
     if _reenter_event_type is not None:
@@ -56,11 +62,21 @@ def register_event_type() -> None:
         raise qtrio.EventTypeRegistrationFailedError()
 
     # assign to the global
-    _reenter_event_type = QtCore.QEvent.Type(event_hint)
+    # TODO: https://bugreports.qt.io/browse/PYSIDE-1347
+    if qts.is_pyqt_5_wrapper:
+        _reenter_event_type = QtCore.QEvent.Type(event_hint)
+    elif qts.is_pyside_5_wrapper:
+        _reenter_event_type = typing.cast(
+            typing.Callable[[int], QtCore.QEvent.Type], QtCore.QEvent.Type
+        )(event_hint)
+    else:  # pragma: no cover
+        raise qtrio.InternalError(
+            "You should not be here but you are running neither PyQt5 nor PySide2.",
+        )
 
 
 def register_requested_event_type(
-    requested_value: typing.Union[int, QtCore.QEvent.Type]
+    requested_value: typing.Union[int, "QtCore.QEvent.Type"]
 ) -> None:
     """Register the requested Qt event type for use by Trio to reenter into the Qt event
     loop.
@@ -75,12 +91,25 @@ def register_requested_event_type(
         qtrio.RequestedEventTypeUnavailableError: if the type returned by Qt does not
             match the requested type.
     """
+    from qts import QtCore
+
     global _reenter_event_type
 
     if _reenter_event_type is not None:
         raise qtrio.EventTypeAlreadyRegisteredError()
 
-    event_hint = QtCore.QEvent.registerEventType(requested_value)
+    # TODO: https://bugreports.qt.io/browse/PYSIDE-1468
+    if qts.is_pyqt_5_wrapper:
+        event_hint = QtCore.QEvent.registerEventType(requested_value)
+    elif qts.is_pyside_5_wrapper:
+        event_hint = typing.cast(
+            typing.Callable[[typing.Union[int, QtCore.QEvent.Type]], int],
+            QtCore.QEvent.registerEventType,
+        )(requested_value)
+    else:  # pragma: no cover
+        raise qtrio.InternalError(
+            "You should not be here but you are running neither PyQt5 nor PySide2.",
+        )
 
     if event_hint == -1:
         raise qtrio.EventTypeRegistrationFailedError()
@@ -90,29 +119,20 @@ def register_requested_event_type(
         )
 
     # assign to the global
-    _reenter_event_type = QtCore.QEvent.Type(event_hint)
+    # TODO: https://bugreports.qt.io/browse/PYSIDE-1347
+    if qts.is_pyqt_5_wrapper:
+        _reenter_event_type = QtCore.QEvent.Type(event_hint)
+    elif qts.is_pyside_5_wrapper:
+        _reenter_event_type = typing.cast(
+            typing.Callable[[int], QtCore.QEvent.Type], QtCore.QEvent.Type
+        )(event_hint)
+    else:  # pragma: no cover
+        raise qtrio.InternalError(
+            "You should not be here but you are running neither PyQt5 nor PySide2.",
+        )
 
 
-class ReenterEvent(QtCore.QEvent):
-    """A proper ``ReenterEvent`` for reentering into the Qt host loop."""
-
-    def __init__(self, fn: typing.Callable[[], object]):
-        super().__init__(_reenter_event_type)
-        self.fn = fn
-
-
-class Reenter(QtCore.QObject):
-    """A ``QtCore.QObject`` for handling reenter events."""
-
-    def event(self, event: QtCore.QEvent) -> bool:
-        """Qt calls this when the object receives an event."""
-
-        reenter_event = typing.cast(Reenter, event)
-        reenter_event.fn()
-        return False
-
-
-async def wait_signal(signal: qtrio._util.SignalInstance) -> typing.Tuple[object, ...]:
+async def wait_signal(signal: "QtCore.SignalInstance") -> typing.Tuple[object, ...]:
     """Block for the next emission of ``signal`` and return the emitted arguments.
 
     Warning:
@@ -158,12 +178,12 @@ class Emission:
         PySide2 or ``QtCore.pyqtBoundSignal`` in PyQt5.
     """
 
-    signal: qtrio._util.SignalInstance
+    signal: "QtCore.SignalInstance"
     """An instance of the original signal."""
     args: typing.Tuple[object, ...]
     """A tuple of the arguments emitted by the signal."""
 
-    def is_from(self, signal: qtrio._util.SignalInstance) -> bool:
+    def is_from(self, signal: "QtCore.SignalInstance") -> bool:
         """Check if this emission came from ``signal``.
 
         Args:
@@ -187,6 +207,24 @@ class Emission:
         return self.is_from(signal=other.signal) and self.args == other.args
 
 
+@attr.s(auto_attribs=True, frozen=True)
+class EmissionsChannelSlot:
+    internal_signal: "QtCore.SignalInstance"
+    send_channel: trio.MemorySendChannel
+
+    def slot(
+        self,
+        *args: object,
+    ) -> None:
+        try:
+            self.send_channel.send_nowait(
+                Emission(signal=self.internal_signal, args=args)
+            )
+        except (trio.WouldBlock, trio.ClosedResourceError):
+            # TODO: log this or... ?
+            pass
+
+
 @attr.s(auto_attribs=True)
 class Emissions:
     """Hold elements useful for the application to work with emissions from signals.
@@ -208,7 +246,7 @@ class Emissions:
 
 @async_generator.asynccontextmanager
 async def open_emissions_channel(
-    signals: typing.Collection[qtrio._util.SignalInstance],
+    signals: typing.Collection["QtCore.SignalInstance"],
     max_buffer_size: typing.Union[int, float] = math.inf,
 ) -> typing.AsyncGenerator[Emissions, None]:
     """Create a memory channel fed by the emissions of the signals.  Each signal
@@ -232,33 +270,26 @@ async def open_emissions_channel(
     # info in a `slot()` stack frame rather than in the memory channel.  Perhaps in the
     # future we can implement a limit beyond which events are thrown away to avoid
     # infinite queueing.  Maybe trio.MemorySendChannel.send_nowait() instead.
-    send_channel, receive_channel = trio.open_memory_channel(
+    send_channel, receive_channel = trio.open_memory_channel[Emission](
         max_buffer_size=max_buffer_size
     )
 
     async with send_channel:
         with contextlib.ExitStack() as stack:
+            emissions = Emissions(channel=receive_channel, send_channel=send_channel)
+
             for signal in signals:
+                slot = EmissionsChannelSlot(
+                    internal_signal=signal, send_channel=send_channel
+                )
+                stack.enter_context(qtrio._qt.connection(signal, slot.slot))
 
-                def slot(
-                    *args: object, internal_signal: qtrio._util.SignalInstance = signal
-                ) -> None:
-                    try:
-                        send_channel.send_nowait(
-                            Emission(signal=internal_signal, args=args)
-                        )
-                    except trio.WouldBlock:
-                        # TODO: log this or... ?
-                        pass
-
-                stack.enter_context(qtrio._qt.connection(signal, slot))
-
-            yield Emissions(channel=receive_channel, send_channel=send_channel)
+            yield emissions
 
 
 @async_generator.asynccontextmanager
 async def enter_emissions_channel(
-    signals: typing.Collection[qtrio._util.SignalInstance],
+    signals: typing.Collection["QtCore.SignalInstance"],
     max_buffer_size: typing.Union[int, float] = math.inf,
 ) -> typing.AsyncGenerator[Emissions, None]:
     """Create a memory channel fed by the emissions of the signals and enter both the
@@ -280,6 +311,32 @@ async def enter_emissions_channel(
                 yield emissions
 
 
+class StarterProtocol(typing_extensions.Protocol):
+    def start(self, *args: object) -> None:
+        ...
+
+
+@attr.s(auto_attribs=True, frozen=True)
+class DirectStarter:
+    slot: typing.Callable[..., typing.Awaitable[object]]
+    nursery: trio.Nursery
+
+    def start(self, *args: object) -> None:
+        self.nursery.start_soon(self.slot, *args)
+
+
+@attr.s(auto_attribs=True, frozen=True)
+class WrappedStarter:
+    slot: typing.Callable[..., typing.Awaitable[object]]
+    wrapper: typing.Callable[
+        [typing.Callable[..., typing.Awaitable[object]]], typing.Awaitable[object]
+    ]
+    nursery: trio.Nursery
+
+    def start(self, *args: object) -> None:
+        self.nursery.start_soon(self.wrapper, self.slot, *args)
+
+
 @attr.s(auto_attribs=True)
 class EmissionsNursery:
     """Holds the nursery, exit stack, and wrapper needed to support connecting signals
@@ -302,26 +359,25 @@ class EmissionsNursery:
 
     def connect(
         self,
-        signal: qtrio._util.SignalInstance,
+        signal: "QtCore.SignalInstance",
         slot: typing.Callable[..., typing.Awaitable[object]],
     ) -> None:
-        """Connect to an async slot to this emissions nursery so when called the slot
+        """Connect an async signal to this emissions nursery so when called the slot
         will be run in the nursery.
         """
-        if self.wrapper is not None:
-
-            def starter(*args: object) -> None:
-                self.nursery.start_soon(self.wrapper, slot, *args)
+        starter: StarterProtocol
 
+        if self.wrapper is None:
+            starter = DirectStarter(slot=slot, nursery=self.nursery)
         else:
+            starter = WrappedStarter(
+                slot=slot, wrapper=self.wrapper, nursery=self.nursery
+            )
 
-            def starter(*args: object) -> None:
-                self.nursery.start_soon(slot, *args)
-
-        self.exit_stack.enter_context(qtrio._qt.connection(signal, starter))
+        self.exit_stack.enter_context(qtrio._qt.connection(signal, starter.start))
 
     def connect_sync(
-        self, signal: qtrio._util.SignalInstance, slot: typing.Callable[..., object]
+        self, signal: "QtCore.SignalInstance", slot: typing.Callable[..., object]
     ) -> None:
         """Connect to a sync slot to this emissions nursery so when called the slot will
         be run in the nursery.
@@ -335,7 +391,7 @@ class EmissionsNursery:
 
 @async_generator.asynccontextmanager
 async def open_emissions_nursery(
-    until: typing.Optional[qtrio._util.SignalInstance] = None,
+    until: typing.Optional["QtCore.SignalInstance"] = None,
     wrapper: typing.Optional[typing.Callable[..., typing.Awaitable[object]]] = None,
 ) -> typing.AsyncGenerator[EmissionsNursery, None]:
     """Open a nursery for handling callbacks triggered by signal emissions.  This allows
@@ -368,7 +424,7 @@ async def open_emissions_nursery(
 
 @async_generator.asynccontextmanager
 async def wait_signal_context(
-    signal: qtrio._util.SignalInstance,
+    signal: "QtCore.SignalInstance",
 ) -> typing.AsyncGenerator[None, None]:
     """Connect a signal during the context and wait for it on exit.  Presently no
     mechanism is provided for retrieving the emitted arguments.
@@ -430,10 +486,10 @@ class Outcomes:
 
 
 def run(
-    async_fn: typing.Callable[[], typing.Awaitable[None]],
+    async_fn: typing.Callable[..., typing.Awaitable[object]],
     *args: object,
     done_callback: typing.Optional[typing.Callable[[Outcomes], None]] = None,
-    clock: trio.abc.Clock = None,
+    clock: typing.Optional[trio.abc.Clock] = None,
     instruments: typing.Sequence[trio.abc.Instrument] = (),
 ) -> object:
     """Run a Trio-flavored async function in guest mode on a Qt host application, and
@@ -474,13 +530,29 @@ def outcome_from_application_return_code(return_code: int) -> outcome.Outcome:
     return outcome.Error(qtrio.ReturnCodeError(return_code))
 
 
-def maybe_build_application() -> QtGui.QGuiApplication:
+def maybe_build_application() -> "QtGui.QGuiApplication":
     """Create a new Qt application object if one does not already exist.
 
     Returns:
         The Qt application object.
     """
-    maybe_application = QtWidgets.QApplication.instance()
+    from qts import QtWidgets  # noqa: F811
+
+    application: QtCore.QCoreApplication
+
+    # TODO: https://bugreports.qt.io/browse/PYSIDE-1467
+    if qts.is_pyqt_5_wrapper:
+        maybe_application = QtWidgets.QApplication.instance()
+    elif qts.is_pyside_5_wrapper:
+        maybe_application = typing.cast(
+            typing.Optional["QtCore.QCoreApplication"],
+            QtWidgets.QApplication.instance(),
+        )
+    else:  # pragma: no cover
+        raise qtrio.InternalError(
+            "You should not be here but you are running neither PyQt5 nor PySide2.",
+        )
+
     if maybe_application is None:
         application = QtWidgets.QApplication(sys.argv[1:])
     else:
@@ -491,11 +563,24 @@ def maybe_build_application() -> QtGui.QGuiApplication:
     return application
 
 
+def create_reenter() -> "qtrio.qt.Reenter":
+    import qtrio.qt
+
+    return qtrio.qt.Reenter()
+
+
+def _early_quit_warning() -> None:
+    warnings.warn(
+        message="The Qt application quit early.  See https://qtrio.readthedocs.io/en/stable/lifetimes.html",
+        category=qtrio.ApplicationQuitWarning,
+    )
+
+
 @attr.s(auto_attribs=True, slots=True)
 class Runner:
     """This class helps run Trio in guest mode on a Qt host application."""
 
-    application: QtGui.QGuiApplication = attr.ib(factory=maybe_build_application)
+    application: "QtGui.QGuiApplication" = attr.ib(factory=maybe_build_application)
     """The Qt application object to run as the host.  If not set before calling
     :meth:`run` the application will be created as
     ``QtWidgets.QApplication(sys.argv[1:])`` and ``.setQuitOnLastWindowClosed(False)``
@@ -506,7 +591,7 @@ class Runner:
     """When true, the :meth:`done_callback` method will quit the application when the
     async function passed to :meth:`qtrio.Runner.run` has completed.
     """
-    clock: trio.abc.Clock = None
+    clock: typing.Optional[trio.abc.Clock] = None
     """The clock to use for this run.  This is primarily used to speed up tests that
     include timeouts.  The value will be passed on to
     :func:`trio.lowlevel.start_guest_run`.
@@ -516,7 +601,7 @@ class Runner:
     :func:`trio.lowlevel.start_guest_run`.
     """
 
-    reenter: Reenter = attr.ib(factory=Reenter)
+    reenter: "qtrio.qt.Reenter" = attr.ib(factory=create_reenter)
     """The :class:`QtCore.QObject` instance which will receive the events requesting
     execution of the needed Trio and user code in the host's event loop and thread.
     """
@@ -540,7 +625,7 @@ class Runner:
 
     def run(
         self,
-        async_fn: typing.Callable[[], typing.Awaitable[object]],
+        async_fn: typing.Callable[..., typing.Awaitable[object]],
         *args: object,
         execute_application: bool = True,
     ) -> Outcomes:
@@ -567,10 +652,13 @@ class Runner:
             args,
             run_sync_soon_threadsafe=self.run_sync_soon_threadsafe,
             done_callback=self.trio_done,
-            clock=self.clock,
+            clock=self.clock,  # type: ignore[arg-type]
             instruments=self.instruments,
         )
 
+        if self.quit_application:
+            self.application.aboutToQuit.connect(_early_quit_warning)
+
         if execute_application:
             return_code = self.application.exec_()
 
@@ -589,7 +677,9 @@ class Runner:
         Args:
             fn: A no parameter callable.
         """
-        event = ReenterEvent(fn=fn)
+        import qtrio.qt
+
+        event = qtrio.qt.ReenterEvent(fn=fn)
         self.application.postEvent(self.reenter, event)
 
     async def trio_main(
@@ -597,8 +687,8 @@ class Runner:
         async_fn: typing.Callable[..., typing.Awaitable[object]],
         args: typing.Tuple[object, ...],
     ) -> object:
-        """Will be run as the main async function by the Trio guest.  It creates a
-        cancellation scope to be cancelled when
+        """Will be run as the main async function by the Trio guest.  If it is a GUI
+        application then it creates a cancellation scope to be cancelled when
         :meth:`QtGui.QGuiApplication.lastWindowClosed` is emitted.  Within this scope
         the application's ``async_fn`` will be run and passed ``args``.
 
@@ -610,11 +700,16 @@ class Runner:
         Returns:
             The result returned by `async_fn`.
         """
+        from qts import QtGui
+
         result: object = None
 
         with trio.CancelScope() as self.cancel_scope:
             with contextlib.ExitStack() as exit_stack:
-                if self.application.quitOnLastWindowClosed():
+                if (
+                    isinstance(self.application, QtGui.QGuiApplication)
+                    and self.application.quitOnLastWindowClosed()
+                ):
                     exit_stack.enter_context(
                         qtrio._qt.connection(
                             signal=self.application.lastWindowClosed,
@@ -644,6 +739,7 @@ class Runner:
             self.done_callback(self.outcomes)
 
         if self.quit_application:
+            self.application.aboutToQuit.disconnect(_early_quit_warning)
             self.application.quit()
 
         self._done = True
